from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    AnnualWellnessVisit, CareServicesInLongTermResidentialFacility,
    DischargeServicesNursingFacility_1065, EncounterInpatient, Ethnicity, HomeHealthcareServices,
    HospiceCareAmbulatory, NursingFacilityVisit, OfficeVisit, OncAdministrativeSex, Payer,
    PneumococcalVaccine, PreventiveCareServicesEstablishedOfficeVisit18AndUp,
    PreventiveCareServicesInitialOfficeVisit18AndUp, Race)


class ClinicalQualityMeasure127v9(ClinicalQualityMeasure):
    """
    Pneumococcal Vaccination Status for Older Adults

    Description: Percentage of patients 65 years of age and older who have ever received a
    pneumococcal vaccine

    Definition: None

    Rationale: Pneumonia is a common cause of illness and death in the elderly and persons with
    certain underlying conditions. The major clinical syndromes of pneumococcal disease include
    pneumonia, bacteremia and meningitis, with pneumonia being the most common (Centers for Disease
    Control and Prevention [CDC], 2015a). Pneumonia symptoms generally include fever, chills,
    pleuritic chest pain, cough with sputum, dyspnea, tachypnea, hypoxia tachycardia, malaise and
    weakness. There are an estimated 400,000 cases of pneumonia in the U.S. each year and a 5
    percent-7 percent mortality rate, although it may be higher among older adults and adults in
    nursing homes (CDC, 2015b; Janssens & Krause, 2004).

    Among the 91.5 million US adults aged > 50 years, 29,500 cases of IPD, 502,600 cases of
    nonbacteremic pneumococcal pneumonia and 25,400 pneumococcal-related deaths are estimated to
    occur yearly; annual direct and indirect costs are estimated to total $3.7 billion and $1.8
    billion, respectively. Pneumococcal disease remains a substantial burden among older US adults,
    despite increased coverage with 23-valent pneumococcal polysaccharide vaccine, (PPV23) and
    indirect benefits afforded by PCV7 vaccination of young children (Weycker et al., 2011).

    Pneumococcal vaccines have been shown to be highly effective in preventing invasive
    pneumococcal disease. When comparing costs, outcomes and quality adjusted life years,
    immunization with the two recommended pneumococcal vaccines was found to be more economically
    efficient than no vaccination, with an incremental cost-effectiveness ratio of $25,841 per
    quality-adjusted life year gained (Chen et al., 2014).

    Guidance: Patient self-report for procedures as well as immunization s should be recorded in
    'Procedure, Performed' template or 'Immunization, Administered' template in QRDA-1.

    ACIP (Kobayashi, 2015) provides guidance about the proper interval and relative timing for the
    administration of two pneumococcal vaccines; this measure assesses whether patients have
    received at least one of either vaccine.

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5. Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms127v9
    """

    title = 'Pneumococcal Vaccination Status for Older Adults'

    identifiers = ['CMS127v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'National Committee for Quality Assurance',
    ]

    references = [
        'Centers for Disease Control and Prevention. (2015a, June 10). Pneumococcal disease: Symptoms and complications. Retrieved from https://www.cdc.gov/pneumococcal/about/symptoms-complications.html',
        'Centers for Disease Control and Prevention. (2015b, June 19). Pneumococcal vaccination: Clinical Features. Retrieved from http://www.cdc.gov/pneumococcal/clinicians/clinical-features.html#pneumonia',
        'Chen, J., O’Brien, M. A., Yang, H. K., et al. (2014). Cost-effectiveness of pneumococcal vaccines for adults in the United States. Advances in Therapy, 31(4), 392-409.',
        'Janssens, J. P., & Krause, K. H. (2004). Pneumonia in the very old. Lancet Infectious Diseases, 4(2), 112-124.',
        'Kobayashi, M., Bennett, N. M., Gierke, R., et al. (2015). "Intervals between PCV13 and PPSV23 vaccines: Recommendations of the Advisory Committee on Immunization Practices (ACIP)." Morbidity and Mortality Weekly Report, 64(34), 947.',
        'Janssens, J.P., and K.H. Krause. 2004. Pneumonia in the very old. Lancet Infect Dis. 4(2):112–24.',
        'National Heart, Lung, and Blood Institute. (2011). "Pneumonia." Retrieved from http://www.nhlbi.nih.gov/health/dci/Diseases/pnu/pnu_whatis.html',
        'Weycker, D., Strutton, D., Edelsberg, J., et al. (2011). "Clinical and economic burden of pneumococcal disease in older U.S. adults." Vaccine, 28(31), 4955-4960.',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: Patients 65 years of age and older with a visit during the measurement
        period
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Exclude patients whose hospice care overlaps the measurement period

        Exceptions: None
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Patients who have ever received a pneumococcal vaccination before the end of the
        measurement period

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: In 2014, the Advisory Committee on Immunization Practices (ACIP)
        began recommending a dose of 13-valent pneumococcal conjugate vaccine (PCV13) be followed
        by a dose of 23-valent pneumococcal polysaccharide vaccine (PPSV23) 6-12 months later in
        adults aged 65 and older who have not previously received a pneumococcal vaccination, and
        in persons over the age of two years who are considered to be at higher risk for
        pneumococcal disease due to an underlying condition. The two vaccines should not be
        coadministered and intervals for administration of the two vaccines vary slightly depending
        on the age, risk group, and history of vaccination (Kobayashi et al., 2015).

        In 2015, ACIP updated its recommendation and changed the interval between PCV13 and PPSV23,
        from 6-12 months to at least one year for immunocompetent adults aged >=65 years who have
        not previously received pneumococcal vaccine. For immunocompromised vaccine-naïve adults,
        the minimum acceptable interval between PCV13 and PPSV23 is 8 weeks. Both immunocompetent
        and immunocompromised adults aged >=65 years who have previously received a dose of PPSV23
        when over the age of 65 should receive a dose of PCV13 at least one year after PPSV23 (>=1
        year). Immunocompetent and immunocompromised adults aged >=65 who have previously received
        a dose of PPSV23 when under the age of 65, should also receive a dose of PCV13 at least one
        year after PPSV23 (>=1 year) and then another dose of PPSV23 at least one year after PCV13.
        It is recommended that for those that have this alternative three-dose schedule (2 PPSV23
        and 1 PCV13), the three doses should be spread over a time period of five or more years
        (Kobayashi et al., 2015).
        """
        pass



"""
Influence Types and Use Case Examples

1. Sorts/Searches: Define choice set sortings under specific conditions
2. Defaults: Set order field value defaults under specific conditions
3. Recommendations: Suggest documentation, orders, analyses, and patient engagements
4. Rationales: Request or require rationales when behavior diverges from protocols
5. Tasks: Assign a task to a user

Below is a small sample of the 200+ events in charting and coordination available for subscription.

Charting Events

CHART_OPENED, CHART_QUERIED, PRESCRIPTION_ORIGINATED, PRESCRIPTION_COMMITTED, REFERRAL_ORIGINATED,
REFERRAL_COMMITTED, NOTE_LOCKED, ...

Care Coordination Events

APPOINTMENT_SCHEDULED, PATIENT_ARRIVED, PATIENT_ROOMED, REFILL_REQUEST_RECEIVED,
REFERRAL_LOOP_CLOSED, POPULATION_QUERIED, MULTIPATIENT_ACTION_ORIGINATED, ...
"""

import importlib
import pkgutil
import re

from collections import defaultdict
from typing import Dict, List, Optional, Type, TypeVar

import arrow
import requests

from cached_property import cached_property
from canvas_workflow_kit import settings

from memoization import cached
from requests.auth import HTTPDigestAuth

from . import events
from .constants import CHANGE_TYPE
from .internal.attrdict import AttrDict
from .internal.string import snakecase
from .metaclass import DeclarativeFieldsMetaclass
from .patient import Patient
from .intervention import Intervention
from .recommendation import (
    Recommendation, LabRecommendation, PlanRecommendation,
    TaskRecommendation, ReferRecommendation, ImagingRecommendation,
    PerformRecommendation, HyperlinkRecommendation, InterviewRecommendation,
    PrescribeRecommendation, VitalSignRecommendation, InstructionRecommendation,
    ImmunizationRecommendation, StructuredAssessmentRecommendation
)
from .timeframe import Timeframe

STATUS_DUE = 'due'
STATUS_SNOOZED = 'snoozed'
STATUS_NOT_APPLICABLE = 'not_applicable'
STATUS_PENDING = 'pending'
STATUS_SATISFIED = 'satisfied'
STATUS_NOT_RELEVANT = 'not_relevant'

PROTOCOL_STATUS_CHOICES = (
    (STATUS_DUE, 'Due'),
    (STATUS_NOT_APPLICABLE, 'Not Applicable'),
    (STATUS_PENDING, 'Pending'),
    (STATUS_NOT_RELEVANT, 'Not Relevant'),
)

# regex to find source URLs in protocol reference data
PROTOCOL_SOURCE_URL_RE = re.compile(r'(?i)(.*)\s((?:https?://)?(?:\S+/\S*))\.?$')

CONTEXT_GUIDANCE = 'guidance'
CONTEXT_REPORT = 'report'

session = requests.Session()

try:
    EXTERNAL_PROTOCOLS_URL = (
        f'{settings.INTEGRATION_SERVICE_ENDPOINT}/retrieve-external-protocols/')

    session.auth = HTTPDigestAuth(settings.INTEGRATION_SERVICE_USERNAME,
                                  settings.INTEGRATION_SERVICE_PASSWORD)

    EXTERNAL_PROTOCOLS_ENABLED = all((
        settings.INTEGRATION_SERVICE_ENDPOINT,
        settings.INTEGRATION_SERVICE_USERNAME,
        settings.INTEGRATION_SERVICE_PASSWORD,
    ))
except:
    EXTERNAL_PROTOCOLS_ENABLED = False
    EXTERNAL_PROTOCOLS_URL = ''


@cached(ttl=60)
def get_external_protocols(protocol_patient):
    """
    {
        "protocols": [
            {
                "identifier" : "CMS131v7"
                "status" : "due",
                "externalNarrative" : "Claims as of Jun 11 2020 suggest this patient has ..."
            },

            ...
        ]
    }
    """
    if not EXTERNAL_PROTOCOLS_ENABLED:
        return []

    response = session.post(EXTERNAL_PROTOCOLS_URL, json=protocol_patient)
    response.raise_for_status()

    response_json = response.json()

    if 'error' in response_json:
        return []

    if 'protocols' not in response_json:
        return []

    return response_json['protocols']


RECOMMENDATION_CLASSES = (
    LabRecommendation, PlanRecommendation,
    TaskRecommendation, ReferRecommendation, ImagingRecommendation,
    PerformRecommendation, HyperlinkRecommendation, InterviewRecommendation,
    PrescribeRecommendation, VitalSignRecommendation, InstructionRecommendation,
    ImmunizationRecommendation, StructuredAssessmentRecommendation
)


def add_recommendation_factory(recommendation_class):
    """
    Generate generic add_*_recommendation methods.
    Eg: add_lab_recommendation(*args, **kwargs)
    """
    def add_recommendation_class(result, *args, **kwargs):
        recommendation = recommendation_class(*args, **kwargs)
        result.recommendations.append(recommendation)
    return add_recommendation_class


class ProtocolResult:
    recommendations: List[Recommendation]
    status: str
    narratives: List[str]
    due_in: Optional[int]
    days_of_notice: int
    next_review: arrow

    def __init__(self):
        self.recommendations = []
        self.status = STATUS_NOT_APPLICABLE
        self.narratives = []
        self.due_in = None
        self.days_of_notice = 30
        self.next_review = None

        # Create add_*_recommendation methods
        for recommendation_class in RECOMMENDATION_CLASSES:
            dynamic_class = add_recommendation_factory(recommendation_class)
            snake_case_type = snakecase(recommendation_class.__name__)
            setattr(self.__class__, f'add_{snake_case_type}', dynamic_class)

    def add_recommendation(self, *args, **kwargs):
        """
        Add a recommendation to the recommendations list.
        """
        if len(args) > 0 and isinstance(args[0], (Recommendation, Intervention)):
            recommendation = args[0]
        else:
            recommendation = Recommendation(*args, **kwargs)

        self.recommendations.append(recommendation)

    def add_narrative(self, narrative: str):
        self.narratives.append(narrative)

    @property
    def narrative(self) -> str:
        return '\n'.join(self.narratives)


class ProtocolSource(object):

    def __init__(self, text: str, url: Optional[str] = None) -> None:
        self.text = text
        self.url = url


class Protocol(metaclass=DeclarativeFieldsMetaclass):

    class Meta:
        title: str = ''
        compute_on_change_types: List[str] = []
        responds_to_event_types: List[str] = []
        identifiers: List[str] = []
        description: str = ''
        information: str = ''
        references: List[str] = []

    def __init__(self, patient: Patient, **kwargs):
        self.patient = patient

    @property
    def identifier(self) -> str:
        identifiers = self._meta.identifiers
        if len(identifiers):
            return identifiers[0]
        return ""

    @classmethod
    def protocol_key(cls) -> str:
        """
        External key used to identify the protocol.
        """
        return cls.__name__

    @staticmethod
    def relative_float(value: str) -> float:
        try:
            sigma = 0.0
            if value[0] == '<' and value[1] != '=':
                sigma = -1e-6
            elif value[0] == '>' and value[1] != '=':
                sigma = +1e-6
            return float(value.strip('<≤=≥>')) + sigma
        except (ValueError, IndexError):
            return 0

    def sources(self) -> List[ProtocolSource]:
        """
        Returns a list of ProtocolSource objects with reference text and URLs. Not all sources have
        URLs.
        """
        sources = []

        for ref in self._meta.references:
            # look for URL at the end of the text
            match = PROTOCOL_SOURCE_URL_RE.search(ref)

            text = ref
            url = None

            if match:
                text = match.group(1)
                url = match.group(2)

            sources.append(ProtocolSource(text, url))

        return sources

    def status(self) -> str:
        raise NotImplementedError('status must be overridden')

    def snoozed(self) -> bool:
        return (self.patient.protocol_overrides.all_switched_off or
                self.patient.protocol_overrides.is_snoozed([self.protocol_key()]))

    @classmethod
    def enabled(cls) -> bool:
        """
        Protocols which return False from this method will be skipped.
        """
        return True


class ClinicalQualityMeasure(Protocol):

    # The data sources that will invalidate this CQM; e.g. if this CQM uses interviews then if a
    # patient interview is committed we will need to re-run this CQM
    #
    # Deprecated - refer to constants.CHANGE_TYPE directly
    CHANGE_ALLERGY_INTOLERANCE = CHANGE_TYPE.ALLERGY_INTOLERANCE
    CHANGE_APPOINTMENT = CHANGE_TYPE.APPOINTMENT
    CHANGE_BILLING_LINE_ITEM = CHANGE_TYPE.BILLING_LINE_ITEM
    CHANGE_CONDITION = CHANGE_TYPE.CONDITION
    CHANGE_COVERAGE = CHANGE_TYPE.COVERAGE
    CHANGE_ENCOUNTER = CHANGE_TYPE.ENCOUNTER
    CHANGE_IMAGING_REPORT = CHANGE_TYPE.IMAGING_REPORT
    CHANGE_IMMUNIZATION = CHANGE_TYPE.IMMUNIZATION
    CHANGE_INSTRUCTION = CHANGE_TYPE.INSTRUCTION
    CHANGE_INTERVIEW = CHANGE_TYPE.INTERVIEW
    CHANGE_LAB_REPORT = CHANGE_TYPE.LAB_REPORT
    CHANGE_MEDICATION = CHANGE_TYPE.MEDICATION
    CHANGE_PATIENT = CHANGE_TYPE.PATIENT
    CHANGE_PROTOCOL_OVERRIDE = CHANGE_TYPE.PROTOCOL_OVERRIDE
    CHANGE_REFERRAL_REPORT = CHANGE_TYPE.REFERRAL_REPORT
    CHANGE_SUSPECT_HCC = CHANGE_TYPE.SUSPECT_HCC
    CHANGE_VITAL_SIGN = CHANGE_TYPE.VITAL_SIGN

    class Meta:
        responds_to_event_types = [
            events.HEALTH_MAINTENANCE,
        ]

    def __init__(self,
                 change_types: List = None,
                 timeframe: Timeframe = None,
                 now: arrow.Arrow = None,
                 context: str = CONTEXT_REPORT,
                 **kwargs):
        super().__init__(**kwargs)

        self.context = context if context == CONTEXT_GUIDANCE else CONTEXT_REPORT
        self.settings = AttrDict()
        self.field_changes = {}

        self._results: Optional[ProtocolResult] = None
        self.canvas_updates = []

        if not self.impacted_by_changes(change_types):
            self._results = ProtocolResult()
            self._results.status = STATUS_NOT_RELEVANT

            return

        self.now = now if now else arrow.utcnow()  # TODO now should be local time

        if timeframe:
            self.timeframe = timeframe
        else:
            end = self.now

            if self.snooze_adjustment:
                start = self.now.shift(days=self.snooze_adjustment['snoozedDays'])
                end = start.shift(years=1)
            elif self.period_adjustment:
                cycle = self.period_adjustment['cycleDays']
                start = end.shift(days=-1 * cycle)
            else:
                start = end.shift(years=-1)  # default duration is one year

            self.timeframe = Timeframe(start=start, end=end)

    def impacted_by_changes(self, change_types: Optional[List]) -> bool:
        # TODO: See if the calling process can select and compute only the protocols
        # that match change-types, instead of instantiating in order to check.
        # Will help reduce code, and speed things up.
        if not change_types or not self._meta.compute_on_change_types:
            return True

        for change in self._meta.compute_on_change_types:
            if change in change_types:
                return True

        return False

    @cached_property
    def snooze_adjustment(self) -> Optional[Dict]:
        override = self.patient.protocol_overrides.defined_for([self.protocol_key()])
        return override['snooze'] if override else None

    @cached_property
    def period_adjustment(self) -> Optional[Dict]:
        override = self.patient.protocol_overrides.defined_for([self.protocol_key()])
        return override['adjustment'] if override else None

    def in_initial_population(self) -> bool:
        """
        Protocols must return a boolean indicating whether a patient
        is in the initial population.
        """
        raise NotImplementedError('in_initial_population must be overridden')

    def in_denominator(self) -> bool:
        """
        Protocols must return a boolean indicating whether a patient
        is in the denominator.
        """
        raise NotImplementedError('in_denominator must be overridden')

    def in_numerator(self) -> bool:
        """
        Protocols must return a boolean indicating whether a patient
        is in the numerator.
        """
        raise NotImplementedError('in_numerator must be overridden')

    def compute_results(self) -> ProtocolResult:
        """
        Protocols must implement this method which returns:

           {
             'status': status,
             'narrative': narrative,
             'recommendations': recommendations
           }

        Status must be one of PROTOCOL_STATUS_CHOICES. Narrative must be an English description of
        recommended actions. Recommendations must be an array of Recommendation sub-class objects
        (LabRecommendation, VaccineRecommendation, etc.) or an empty array.
        """
        raise NotImplementedError('compute_results must be overridden')

    def results(self) -> ProtocolResult:
        """
        Calls compute_results and caches the result.
        """
        if not self._results:
            self._results = self.compute_results()

        return self._results

    def set_context(self, context: Dict = {}) -> None:
        """
        Deprecated - using to populate settings and field changes for pre-2022-03 instances
        Use set_settings and
        """
        self.settings = AttrDict(context.get('config', {}))
        self.field_changes = context.get('change_info', {})

    def set_settings(self, settings: Dict = {}) -> None:
        self.settings = AttrDict(settings)

    def set_updates(self, canvas_updates: List[Dict]) -> None:
        self.canvas_updates = canvas_updates

    def updates(self) -> List[Dict]:
        return self.canvas_updates

    def recommendations(self) -> List[Recommendation]:
        return self.results().recommendations

    def narrative(self) -> str:
        return self.results().narrative

    def status(self) -> str:
        return self.results().status

    def due_in(self) -> Optional[int]:
        return self.results().due_in

    def days_of_notice(self) -> int:
        return self.results().days_of_notice

    def next_review(self) -> arrow:
        next_review = self.results().next_review
        due_in = self.results().due_in
        if due_in and due_in > 0 and next_review is None:
            next_review = arrow.utcnow().shift(days=due_in)
        return next_review

    def display_date(self, day: arrow.Arrow) -> str:
        return '%s on %s' % (day.humanize(other=self.now), day.format('M/D/YY'))

    @staticmethod
    def display_period(day_from: arrow.Arrow, day_to: arrow.Arrow) -> str:
        return 'between %s and %s' % (day_from.format('M/D/YY'), day_to.format('M/D/YY'))

    def screening_interval_context(self) -> str:
        try:
            abc = self.friendly_time_duration(self._meta.default_display_interval_in_days)
        except (AttributeError, KeyError):
            return ''

        if self.period_adjustment:
            abc = self.friendly_time_duration(self.period_adjustment['cycleDays'])
        return f'Current screening interval {abc}.'

    @staticmethod
    def friendly_time_duration(duration_in_days: int) -> str:
        friendly_duration = 'invalid duration'
        if duration_in_days >= 1:
            friendly_duration = f'{duration_in_days} days'
            if duration_in_days >= 365:
                years, days = divmod(duration_in_days, 365)
                plural = 's' if years > 1 else ''
                friendly_duration = f'{years} year{plural}'
                if days >= 30:
                    months = days // 30
                    plural = 's' if months > 1 else ''
                    friendly_duration = friendly_duration + f', {months} month{plural}'
            elif duration_in_days >= 30:
                months, days = divmod(duration_in_days, 30)
                plural = 's' if months > 1 else ''
                friendly_duration = f'{months} month{plural}'
                if days > 0:
                    plural = 's' if days > 1 else ''
                    friendly_duration = friendly_duration + f', {days} day{plural}'
        return friendly_duration

    @staticmethod
    def group_question_responses(interview: Dict) -> Dict:
        result: Dict = {}
        qr_id_2_code: Dict = {}

        for question in interview.get("questions", []):
            qr_id_2_code[question['questionResponseId']] = question['code']

        for response in interview.get('responses', []):
            question_code = qr_id_2_code[response['questionResponseId']]
            if question_code not in result:
                result[question_code] = []
            result[question_code].append(response['code'])

        return result

    # TODO move to ClinicalQualityMeasureReport class
    # def report(self):
    #     """
    #     Report aggregate statistical information on patients who are included or excluded from
    #     this protocol's criteria.
    #     """
    #     return {}

    # for Anthem we only actually know if a patient is in both the denominator AND NOT the
    # numerator, not both separately

    def format_external_protocols(self):
        return {
            'patientData': {
                'firstName': self.patient.first_name,
                'lastName': self.patient.last_name,
                'birthDate': self.patient.date_of_birth,
                'coverages': self.patient.coverages,
            }
        }


class ExternallyAwareClinicalQualityMeasure:

    @property
    def external_protocol(self):
        if not EXTERNAL_PROTOCOLS_ENABLED:
            return {'externalNarrative': '', 'status': None}

        external_protocols = get_external_protocols(self.format_external_protocols())

        for protocol in external_protocols:
            if protocol['identifier'] == self.identifier:
                return protocol

        return {'externalNarrative': '', 'status': None}

    @property
    def external_narrative(self):
        return self.external_protocol['externalNarrative']

    def in_external_denominator(self):
        return self.external_protocol['status'] in ('due', 'satisfied')

    def in_external_numerator(self):
        return self.external_protocol['status'] == 'satisfied'

    # send soft close, since we think it is satisfied and external parties do
    # not
    def send_soft_close(self):
        # print('SOFT CLOSE: REPLACE ME WITH A CALL TO EXTERNAL PROTOCOL UPDATE ENDPOINT')
        pass

    # canvas does not think it applies, external does
    def applicable_externally_but_not_in_canvas(self):
        # for now, do nothing. In the future, if we have appropriate external
        # clinical context, try to capture the context within canvas.
        # Example: "claims suggest patient was diagnosed with xxx by xxx on
        # date xxx."
        # Options:
        # - Insert Past Medical History
        # - Request Records
        pass

    # protocol does not apply
    def craft_not_applicable_result(self):
        result = ProtocolResult()
        result.due_in = self.first_due_in()
        return result

    def compute_results(self) -> ProtocolResult:
        result: ProtocolResult

        if self.in_denominator():
            if self.in_numerator():
                result = self.craft_satisfied_result()

                if not self.in_external_numerator():
                    self.send_soft_close()
            else:
                result = self.craft_unsatisfied_result()

                if self.in_external_numerator():
                    # protocol satisfied externally, not in canvas
                    # talk to patient, request records, etc.
                    result.add_recommendation(
                        TaskRecommendation(
                            key=f'{self.identifier}_RECOMMEND_RECORDS_REQUEST',
                            button='Request Records',
                            patient=self.patient,
                            narrative=('An insurer believes this protocol has already been '
                                       'satisfied elsewhere. Request that records be retrieved to '
                                       'support this claim.'),
                            title='Request Records'))
                else:
                    # protocol unsatisfied in canvas and externally
                    # if external_narrative, use that narrative, otherwise our narrative
                    if self.external_narrative:
                        result.narratives = [self.external_narrative]
        else:
            result = self.craft_not_applicable_result()

            # if self.in_external_denominator():
            #     canvas does not think it applies, external does
            #     result = self.applicable_externally_but_not_in_canvas()

        return result


T = TypeVar('T')


def get_subclasses(cls: Type[T]) -> List[Type[T]]:
    subclasses = []

    for subclass in cls.__subclasses__():
        subclasses.append(subclass)
        subclasses.extend(get_subclasses(subclass))

    return subclasses


def get_protocols(module) -> Dict[str, List[Type[Protocol]]]:
    for _, name, _ in pkgutil.iter_modules(module.__path__):
        importlib.import_module(f'{module.__name__}.{name}')

    subclasses = get_subclasses(Protocol)
    protocols_by_event_type: Dict[str, List[Type[Protocol]]] = defaultdict(list)

    for subclass in subclasses:
        # Ignore protocol subclasses that are not under the given module.
        if not subclass.__module__.startswith(module.__name__):
            continue

        # don't add the base classes themselves
        if subclass in (Protocol, ClinicalQualityMeasure):
            continue

        if not subclass.enabled():
            continue

        for event_type in subclass._meta.responds_to_event_types:
            protocols_by_event_type[event_type].append(subclass)

    return protocols_by_event_type


def protocols_for_patient(protocols: List[Type[Protocol]],
                          patient: Patient,
                          change_types: List = None,
                          context: str = CONTEXT_REPORT) -> List[ClinicalQualityMeasure]:
    """
    Return the set of protocols for this patient+event_type where the
    protocol is active for the patient and the protocol status is something
    other than Not Applicable.
    """

    def useful_cqm(cqm: ClinicalQualityMeasure) -> bool:
        if cqm.status() == STATUS_NOT_APPLICABLE:
            return bool(cqm.due_in())
        return True

    return [
        p for p in [
            protocol(patient=patient, change_types=change_types, context=context)
            for protocol in protocols
        ] if (isinstance(p, ClinicalQualityMeasure) and useful_cqm(p))
    ]


class CodingStruct:

    def __init__(self, code=None, system=None):
        self.code = code
        self.system = system



import arrow

from canvas_workflow_kit import settings

from canvas_workflow_kit.builtin_cqms.cms122v6_diabetes_hemoglobin_a1c_poor_control import (
    ClinicalQualityMeasure122v6
)
from canvas_workflow_kit.patient_recordset import (
    BillingLineItemRecordSet,
    LabReportRecordSet,
    ProtocolOverrideRecordSet
)
from canvas_workflow_kit.tests.base import SDKBaseTest
from canvas_workflow_kit.timeframe import Timeframe


class ClinicalQualityMeasure122v6Test(SDKBaseTest):

    def setUp(self):
        self.mocks_path = f'{settings.BASE_DIR}/builtin_cqms/tests/mock_data'

    def test_description(self):
        expected = (
            'Patients 18-75 years of age with diabetes who have either a hemoglobin A1c > 9.0% '
            'or no hemoglobin A1c within the last year.')
        self.assertEqual(expected, ClinicalQualityMeasure122v6._meta.description)

    def test_information(self):
        expected = 'https://ecqi.healthit.gov/sites/default/files/ecqm/measures/CMS122v6.html'
        self.assertEqual(expected, ClinicalQualityMeasure122v6._meta.information)

    def test_enabled(self):
        self.assertTrue(ClinicalQualityMeasure122v6.enabled())

        def test_version(self):
            self.assertTrue(hasattr(ClinicalQualityMeasure122v6._meta, 'version'))

    def test_change_types(self):
        result = ClinicalQualityMeasure122v6._meta.compute_on_change_types
        expected = [
            'protocol_override',
            'condition',
            'lab_report',
            'patient',
        ]
        self.assertEqual(len(expected), len(result))
        for change in expected:
            self.assertIn(change, result)

    def test_last_hba1c_record(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient has diabetes --> true
        patient = self.load_patient('cms122v6_diabetes_yeswithtest')
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe)
        self.assertTrue('id' in tested.last_hba1c_record)
        self.assertEqual(644, tested.last_hba1c_record['id'])

        # patient has diabetes + already perform the exam --> true
        patient = self.load_patient('cms122v6_diabetes_yesnotest')
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe)
        self.assertIsNone(tested.last_hba1c_record)

    def test_in_numerator_with_override(self):
        patient = self.load_patient('cms122v6_diabetes_yeswithtest')
        patient.protocol_overrides = ProtocolOverrideRecordSet([{
            'protocolKey': 'ClinicalQualityMeasure122v6',
            'adjustment': {
                'reference': '2018-02-01T00:00:00Z',
                'cycleDays': 180
            },
            'snooze': None,
            'modified': '2018-03-15T12:32:54Z',
        }])  # yapf: disable
        patient.lab_reports[0]['originalDate'] = '2018-08-05'

        tested = ClinicalQualityMeasure122v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=0))
        self.assertTrue('id' in tested.last_hba1c_record)
        self.assertEqual(644, tested.last_hba1c_record['id'])

        tested = ClinicalQualityMeasure122v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=80))
        self.assertTrue('id' in tested.last_hba1c_record)
        self.assertEqual(644, tested.last_hba1c_record['id'])

        tested = ClinicalQualityMeasure122v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=180))
        self.assertTrue('id' in tested.last_hba1c_record)
        self.assertEqual(644, tested.last_hba1c_record['id'])

        tested = ClinicalQualityMeasure122v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=181))
        self.assertIsNone(tested.last_hba1c_record)

    def test_last_hba1c_value(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient has diabetes --> true
        patient = self.load_patient('cms122v6_diabetes_yeswithtest')
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe)
        self.assertEqual(7, tested.last_hba1c_value)

        # patient has diabetes + already perform the exam --> true
        patient = self.load_patient('cms122v6_diabetes_yesnotest')
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe)
        self.assertIsNone(tested.last_hba1c_value)

        # patient with invalid value --> 0
        patient = self.load_patient('cms122v6_diabetes_yesnotest')
        patient.lab_reports = LabReportRecordSet([
            {
                'originalDate': '2018-08-02',
                'loincCodes': [{
                    'code': '17856-6',
                    'value': 644
                }],
                'value': 'asdf',
            },
        ])
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe)
        self.assertEqual(0, tested.last_hba1c_value)

    def test_last_hba1c_date(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient has diabetes --> true
        patient = self.load_patient('cms122v6_diabetes_yeswithtest')
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe)
        text = r'^[\w\s]+ on 8/2/18$'
        self.assertRegex(tested.last_hba1c_date, text)

        # patient has diabetes + already perform the exam --> true
        patient = self.load_patient('cms122v6_diabetes_yesnotest')
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe)
        self.assertIsNone(tested.last_hba1c_date)

    def test_in_denominator(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient has no diabetes --> false
        patient = self.load_patient('cms122v6_diabetes_no')
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

        # patient has diabetes --> true
        patient = self.load_patient('cms122v6_diabetes_yeswithtest')
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

        # -- patient to young < 18 --> false
        patient = self.load_patient('cms122v6_diabetes_yeswithtest')
        patient.patient['birthDate'] = '2000-08-24'
        self.assertTrue(17 < patient.age_at(timeframe.end) < 18)
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())
        patient.patient['birthDate'] = '2000-08-23'
        self.assertTrue(patient.age_at(timeframe.end) == 18)
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

        # -- patient to old > 75 --> false
        patient = self.load_patient('cms122v6_diabetes_yeswithtest')
        patient.patient['birthDate'] = '1943-08-22'
        self.assertTrue(75 < patient.age_at(timeframe.end) < 76)
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())
        patient.patient['birthDate'] = '1943-08-23'
        self.assertTrue(patient.age_at(timeframe.end) == 75)
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())
        patient.patient['birthDate'] = '1943-08-24'
        self.assertTrue(patient.age_at(timeframe.end) < 75)
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

        # -- patient has an other code --> false
        patient = self.load_patient('cms122v6_diabetes_yeswithtest')
        for condition in patient.conditions:
            for code in condition['coding']:
                code['code'] = 'code'
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

        # patient has diabetes + already perform the exam --> true
        patient = self.load_patient('cms122v6_diabetes_yeswithtest')
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

    def test_in_numerator(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # not test when patient has no diabetes because it makes no sense
        # patient = self.load_patient('cms122v6_diabetes_no')

        # patient has diabetes + exam performed with value under 9 --> false
        patient = self.load_patient('cms122v6_diabetes_yeswithtest')
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())

        # patient has diabetes + exam performed with value equals to 9 --> false
        patient = self.load_patient('cms122v6_diabetes_yeswithtest')
        patient.lab_reports.records[0]['value'] = 9
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())

        # patient has diabetes + exam performed with value above 9 --> true
        patient = self.load_patient('cms122v6_diabetes_yeswithtest')
        patient.lab_reports.records[0]['value'] = 10
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())

        # patient has diabetes + no exam --> true
        patient = self.load_patient('cms122v6_diabetes_yesnotest')
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())

    def test_compute_results(self):
        start = arrow.get('2017-10-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        now = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient has no diabetes --> n/a
        patient = self.load_patient('cms122v6_diabetes_no')
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertIsNotApplicable(result)

        # patient has diabetes + exam performed with value under 9 --> satisfied
        patient = self.load_patient('cms122v6_diabetes_yeswithtest')
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('satisfied', result.status)
        text = r"^Dohav's last HbA1c done [\w\s]+ on 8/2/18 was 7.0%.$"
        self.assertRegex(result.narrative, text)
        self.assertEqual([], result.recommendations)
        self.assertEqual(282, result.due_in)  # not one year since the time frame define the period

        # patient has diabetes + exam performed with value above 9 --> due
        patient = self.load_patient('cms122v6_diabetes_yeswithtest')
        patient.lab_reports.records[0]['value'] = 11.23
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        text = r"^Dohav's last HbA1c done [\w\s]+ on 8/2/18 was 11.2%.$"
        self.assertRegex(result.narrative, text)
        expected = [{
            'title': (
                'Discuss lifestyle modification and medication adherence.'
                ' Consider diabetes education and medication intensification as appropriate.'),
            'command': {
                'type': 'instruct',
                'filter': {
                    'coding': [
                        {
                            'system': 'hcpcs',
                            'code': ['S9452', 'S9470']
                        }, {
                            'system': 'icd10cm',
                            'code': ['Z713']
                        }, {
                            'system': 'icd9cm',
                            'code': ['V653']
                        },
                        {
                            'system': 'snomedct',
                            'code': [
                                '304491008', '410177006', '418995006', '410270001', '182922004',
                                '183065007', '370847001', '386464006', '182955009', '182960008',
                                '182954008', '410171007', '306163007', '284352003', '424753004',
                                '103699006', '182956005', '281085002', '183071001', '183061003',
                                '11816003', '61310001', '410114009', '183070000', '443288003',
                                '361231003', '413315001', '284071006'
                            ]
                        }
                    ]
                }
            },
            'key': 'CMS122v6_RECOMMEND_DISCUSS_LIFESTYLE',
            'rank': 1,
            'button': 'Instruct',
        }]
        # self.assertEqual(recommendations, result.recommendations)
        self.helper_compare_recommendations(expected, result.recommendations)
        self.assertEqual(-1, result.due_in)

        # patient has diabetes + no exam --> due
        patient = self.load_patient('cms122v6_diabetes_yesnotest')
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        text = "Dohav's last HbA1c test was over 10 months."
        self.assertEqual(text, result.narrative)
        expected = [
            {
                'title': 'Order HbA1c',
                'command': {
                    'type': 'labOrder',
                    'filter': {
                        'coding': [{
                            'system': 'loinc',
                            'code': ['4548-4', '17856-6', '4549-2']
                        }]
                    }
                },
                'key': 'CMS122v6_RECOMMEND_HBA1C',
                'rank': 1,
                'button': 'Order',
            },
        ]
        self.helper_compare_recommendations(expected, result.recommendations)
        self.assertEqual(-1, result.due_in)

        # patient has diabetes + younger than 18 --> n/a + due in
        patient = self.load_patient('cms122v6_diabetes_no')
        patient.patient['birthDate'] = '2000-08-30'
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('not_applicable', result.status)
        self.assertIsNone(result.due_in)
        patient = self.load_patient('cms122v6_diabetes_yesnotest')
        patient.patient['birthDate'] = '2000-08-30'
        tested = ClinicalQualityMeasure122v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('not_applicable', result.status)
        self.assertEqual(6, result.due_in)

    def test_compute_results_with_override(self):
        patient = self.load_patient('cms122v6_diabetes_yeswithtest')
        patient.protocol_overrides = ProtocolOverrideRecordSet([{
            'protocolKey': 'ClinicalQualityMeasure122v6',
            'adjustment': {
                'reference': '2018-02-01T00:00:00Z',
                'cycleDays': 180
            },
            'snooze': None,
            'modified': '2018-03-15T12:32:54Z',
        }])  # yapf: disable
        patient.lab_reports[0]['originalDate'] = '2018-08-05'

        # exam done within the last 180 days
        tested = ClinicalQualityMeasure122v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=80))
        result = tested.compute_results()
        self.assertEqual('satisfied', result.status)
        text = "Dohav's last HbA1c done 2 months ago on 8/5/18 was 7.0%."
        self.assertEqual(text, result.narrative)
        self.assertEqual(0, len(result.recommendations))
        self.assertEqual(100, result.due_in)

        # exam done 181 days
        patient.billing_line_items = BillingLineItemRecordSet([{
            'id': 92,
            'created': arrow.get('2018-08-01').shift(days=181),
            'datetimeOfService': arrow.get('2018-08-01').shift(days=181),
            'cpt': '99202',
            'units': 1
        }])
        tested = ClinicalQualityMeasure122v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=181))
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        text = "Dohav's last HbA1c test was over 6 months."
        self.assertEqual(text, result.narrative)
        self.assertEqual(1, len(result.recommendations))
        self.assertEqual(-1, result.due_in)

    def helper_compare_recommendations(self, expected, result):
        for exp, recommendation in zip(expected, result):
            self.assertEqual(exp['key'], recommendation.key)
            self.assertEqual(exp['rank'], recommendation.rank)
            self.assertEqual(exp['button'], recommendation.button)
            self.assertEqual(exp['title'], recommendation.title)
            self.assertEqual(exp['command']['type'], recommendation.command['type'])
            exp_item = exp['command']['filter']['coding']
            rec_item = recommendation.command['filter']['coding']
            self.assertEqual(len(exp_item), len(rec_item))
            for exp_coding, rec_coding in zip(exp_item, rec_item):
                self.assertEqual(exp_coding['system'], rec_coding['system'])
                self.assertEqual(sorted(exp_coding['code']), sorted(rec_coding['code']))



# flake8: noqa
from canvas_workflow_kit import events
from canvas_workflow_kit.protocol import ClinicalQualityMeasure
from canvas_workflow_kit.recommendation import ImmunizationRecommendation  # noqa

# context + prompting is huge win (for dignity); e.g. had this vaccine 11 months ago, he's now due
# for pneuomax 23; prevent need to look back in the chart; more trust == more utility
#


class PneuomoniaVaccinationStatusForOlderAdultsProtocol(ClinicalQualityMeasure):
    """
    https://ecqi.healthit.gov/ep/ecqms-2018-performance-period/pneumococcal-vaccination-status-older-adults

    Description: Percentage of patients 65 years of age and older who have ever received a
    pneumococcal vaccine.

    Initial population and denominator: Patients 65 years of age and older with a visit during the
    measurement period.

    Numerator: Patients who have ever received a pneumococcal vaccination.

    Guidance: It is recommended that patients 65 years of age or older receive one pneumococcal
    vaccination in their lifetime.

    The CDC flowchart specifies additional guidance used for narrative generation:

        https://www.cdc.gov/vaccines/vpd/pneumo/downloads/pneumo-vaccine-timing.pdf
    """

    # this protocol was created before an overhaul and needs finishing to work
    @classmethod
    def enabled(cls):
        return False

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    # where does the action go? in this case will it be a Recommendation?

    # request from clientside JS, e.g. for preventative services; figure out a way to serialize and
    # ship to client; include command & parameters (e.g. immunize, CVX 33)
    # need URLs/views for this

    # 'conceptually a protocol is a recommendation'
    # 'not enough for us to just report on whether you've had one of these vaccines; we'll give you
    # specific advice & tell you why'

    # ontologies to use? e.g. CVX 33/CVX 133
    # andrew: use CVX
    # look up value set for these; these will come from VSAC; in cases where there isn't one we can
    # create our own with the same structure
    # pneumococcal_vaccines = ['PCV13', 'PPSV23']
    pneumococcal_vaccines = ['CVX33', 'CVX133']

    # TODO Recommendations can have >1 action (more than one way to resolve a Recommendation)
    # TODO Add Predicate class for rich "debugging"?
    # TODO action-oriented 'tell me what to do view' vs. inverted 'tell me why' view
    # TODO look through dignity pcp/interop/customers/dignityhealth/...
    # https://github.com/canvas-medical/pcp/tree/master/pcp/interop/customers/dignityhealth/allscripts_touchworks/automata
    # TODO soup to nuts on 'kiddo' example

    @staticmethod
    def condition_names(conditions):
        return ', '.join(condition.name for condition in conditions)

    def recommendations(self):
        pvc13 = self.has_vaccine('PVC13')
        ppsv23 = self.has_vaccine('PPSV23')

        pvc13_age = pvc13.age_at_administration
        ppsv23_age = ppsv23.age_at_administration

        first_name = self.patient.first_name

        if pvc13_age >= 65 and ppsv23_age >= 65:
            return [{
                'type': 'up-to-date',
                'narrative': (
                    f'{first_name} received both PVC13 and PPSV23 after the age of 65 (they '
                    f'received PVC13 at {pvc13_age} and PPSV23 at {ppsv23_age}).')
            }]

        narratives = []

        compromised_conditions = (self.is_asplenic_or_immunocompromised() +
                                  self.is_immunocompetent_with_cochlear_implant_or_csf())
        competent_conditions = self.is_immunocompetent_with_condition()

        compromised_names = self.condition_names(compromised_conditions)
        competent_names = self.condition_names(competent_conditions)  # noqa

        # TODO this logic may have missing edge cases; needs to be verified with a test matrix that
        # follows the CDC flowcharts
        if compromised_conditions and not pvc13:
            narratives.append({
                'type': 'due',
                'action': 'vaccination',
                'code': 'PVC13',
                'narrative': (
                    'Patients 65 years or older with certain medical conditions should be '
                    'vaccinated with PVC13 only if they have not had a prior PVC13 vaccination. '
                    f'{first_name} has a qualifying condition ({compromised_names}) and has not '
                    'previously been vaccinated with PVC13.')
            })
        elif pvc13_age < 65:
            narratives.append({
                'type': 'due',
                'action': 'vaccination',
                'code': 'PVC13',
                'narrative': (
                    'Patients 65 years or older should receive a PVC13 vaccination regardless of '
                    'prior PVC13 vaccination status.')
            })

        # "PCV13 and PPSV23 should not be administered during the same office visit."
        # "When both are indicated, PCV13 should be given before PPSV23 whenever possible."
        if any(narrative.get('code') == 'PVC13' for narrative in narratives):
            return narratives

        if not ppsv23:
            if compromised_conditions and pvc13.administered.within('8 weeks'):
                narratives.append({
                    'type': 'up-to-date',
                    'narrative': (
                        f'{first_name} received PVC13 within the last 8 weeks. Because of '
                        f'their qualifying condition(s) ({compromised_names}) they must wait '
                        f"until {pvc13.administered.add('8 weeks')} to receive PPSV23.")
                })
            elif pvc13.administered.within('1 year'):
                narratives.append({
                    'type': 'up-to-date',
                    'narrative': (
                        f'{first_name} received PVC13 within the last year. They must wait '
                        f"until {pvc13.administered.add('1 year')} to receive PPSV23.")
                })
            else:
                narratives.append({
                    'type': 'due',
                    'action': 'vaccination',
                    'code': 'PPSV23',
                    'narrative': ('Patients 65 years or older should receive a PPSV23 vaccination '
                                  'regardless of prior PPSV23 vaccination status.')
                })
        elif ppsv23_age < 65:
            if ((compromised_conditions or competent_conditions) and
                    ppsv23.administered.within('5 years')):
                narratives.append({
                    'type': 'up-to-date',
                    'narrative': (
                        f'{first_name} received PPSV23 before they turned 65 but within the last '
                        f"5 years. They must wait until {ppsv23.administered.add('5 years')} to "
                        'receive PPSV23.')
                })

        return narratives

    def is_immunocompetent_with_condition(self):
        # TODO replace these with value sets (if available) or arrays
        return self.patient.has_any_condition([
            'alcoholism',
            # "Including congestive heart failure and cardiomyopathies"
            'chronic heart disease',
            'chronic liver disease',
            # "Including chronic obstructive pulmonary disease, emphysema, and asthma"
            'chronic lung disease',
            'smoker',
            'diabetes mellitus',
        ])

    def is_immunocompetent_with_cochlear_implant_or_csf(self):
        return self.patient.has_any_condition([
            'cochlear implant',
            'cerebrospinal fluid leak',
        ])

    def is_asplenic_or_immunocompromised(self):
        return self.patient.has_any_condition([
            # asplenias
            'congenital or acquired asplenia',
            'sickle cell disease',
            'other hemoglobinopathies',

            # immunocompromised
            'chronic renal failure',
            # "Includes B- (humoral) or T-lymphocyte deficiency, complement deficiencies
            #  (particularly C1, C2, C3, and C4 deficiencies), and phagocytic disorders (excluding
            #  chronic granulomatous disease)"
            'congenital or acquired immunodeficiencies',
            'generalized malignancy',
            'HIV infection',
            'Hodgkin disease',
            # "Diseases requiring treatment with immunosuppressive drugs, including long-term
            #  systemic corticosteroids and radiation therapy"
            'latrogenic immunosupression',
            'Leukemia',
            'multiple myeloma',
            'nephrotic syndrome',
            'solid organ transplant',
        ])

    # it would be great to be able to specify filters here in the same way that the Django ORM does
    # to avoid writing one test method for a single patient and one for reporting statistics on a
    # number of patients... 'point of care vs. reporting'
    #
    # for reporting also need to handle age during timeframe vs. age at report generation time
    def in_denominator(self):
        # TODO need to exclude patients on hospice care during the timeframe
        return self.patient.age >= 65 and self.patient.has_visit_within(self.timeframe)

    # for reporting need to check if they had pneumococcal_vaccines before the end of the reporting
    # period
    def in_numerator(self):
        return self.patient.vaccines_within(self.timeframe, self.pneumococcal_vaccines)



class Population:

    def __init__(self):
        self.in_initial_population: bool = None
        self.in_denominator: bool = None
        self.in_numerator: bool = None

    def set_initial_population(self, flag: bool):
        self.in_initial_population = flag
        self.in_denominator = flag
        self.in_numerator = flag

    def set_denominator(self, flag: bool):
        self.in_denominator = flag
        self.in_numerator = flag

    def set_numerator(self, flag: bool):
        self.in_numerator = flag



from typing import List, Optional, Union

from canvas_workflow_kit import events
from canvas_workflow_kit.protocol import ClinicalQualityMeasure
from canvas_workflow_kit.timeframe import Timeframe
from canvas_workflow_kit.recommendation import VitalSignRecommendation, InstructionRecommendation, Recommendation

# flake8: noqa
from canvas_workflow_kit.value_set.v2018 import (
    AboveNormalFollowUp, AboveNormalMedications, BelowNormalFollowUp, BelowNormalMedications,
    BmiEncounterCodeSet, BmiLoincValue, Ethnicity, OncAdministrativeSex, PalliativeCareEncounter,
    PalliativeCare, Payer, PregnancyDx, Race, ReferralsWhereWeightAssessmentMayOccur,
    BmiPercentile, Height, Weight)


class ClinicalQualityMeasure69v6(ClinicalQualityMeasure):
    """
    Preventive Care and Screening: Body Mass Index (BMI) Screening and Follow-Up Plan

    Description: Percentage of patients aged 18 years and older with a BMI documented during the
    current encounter or during the previous twelve months AND with a BMI outside of normal
    parameters, a follow-up plan is documented during the encounter or during the previous twelve
    months of the current encounter

    Normal Parameters:       Age 18 years and older BMI => 18.5 and < 25 kg/m2

    Definition: BMI- Body mass index (BMI) is a number calculated using the Quetelet index: weight
    divided by height squared (W/H2) and is commonly used to classify weight categories. BMI can be
    calculated using:

    Metric Units:  BMI = Weight (kg) / (Height (m) x Height (m))
    OR
    English Units: BMI = Weight (lbs.) / (Height (in) x Height (in)) x 703

    Follow-Up Plan - Proposed outline of treatment to be conducted as a result of a BMI out of
    normal parameters. A follow-up plan may include, but is not limited to: documentation of
    education, referral (for example a registered dietician, nutritionist, occupational therapist,
    physical therapist, primary care provider, exercise physiologist, mental health professional,
    or surgeon), pharmacological interventions, dietary supplements, exercise counseling or
    nutrition counseling.

    Rationale: BMI Above Normal Parameters

    Obesity is a chronic, multifactorial disease with complex psychological, environmental (social
    and cultural), genetic, physiologic, metabolic and behavioral causes and consequences. The
    prevalence of overweight and obese people is increasing worldwide at an alarming rate in both
    developing and developed countries. Environmental and behavioral changes brought about by
    economic development, modernization and urbanization have been linked to the rise in global
    obesity. The health consequences are becoming apparent (ICSI 2013. p.6).

    Nationally, nearly 38 percent of adults are obese [NHANES, 2013-2014 data]. Nearly 8 percent of
    adults are extremely obese (BMI greater than or equal to 40.0); Obesity rates are higher among
    women (40.4 percent) compared to men (35.0 percent). Between 2005 and 2014, the difference in
    obesity among women was 5.1 percent higher among women and 1.7 percent higher among men. Women
    are also almost twice as likely (9.9 percent) to be extremely obese compared to men (5.5
    percent); In addition, rates are the highest among middle-age adults (41 percent for 40- to
    59-year-olds), compared to 34.3 percent of 20- to 39-year-olds and 38.5 percent of adults ages
    60 and older (Flegal KM, Kruszon-Moran D, Carroll MD, et al, 2016, p.2286-2290). Obesity is one
    of the biggest drivers of preventable chronic diseases and healthcare costs in the United
    States. Currently, estimates for these costs range from $147 billion to nearly $210 billion per
    year Cawley J and Meyerhoefer C., 2012 & Finkelstein, Trogdon, Cohen, et al., 2009). There are
    significant racial and ethnic inequities [NHANES, 2013-2014 data]: Obesity rates are higher
    among Blacks (48.4 percent) and Latinos (42.6 percent) than among Whites (36.4 percent) and
    Asian Americans (12.6 percent).The inequities are highest among women: Blacks have a rate of
    57.2 percent, Latinos of 46.9 percent, Whites of 38.2 percent and Asians of 12.4 percent. For
    men, Latinos have a rate of 37.9 percent, Blacks of 38.0 percent and Whites of 34.7 percent.
    Black women (16.8 percent) are twice as likely to be extremely obese as White women (9.7
    percent) (Flegal KM, Kruszon-Moran D, Carroll MD, et al., 2016, pp. 2284-2291).

    BMI continues to be a common and reasonably reliable measurement to identify overweight and
    obese adults who may be at an increased risk for future morbidity. Although good quality
    evidence supports obtaining a BMI, it is important to recognize it is not a perfect
    measurement. BMI is not a direct measure of adiposity and as a consequence it can over- or
    underestimate adiposity. BMI is a derived value that correlates well with total body fat and
    markers of secondary complications, e.g., hypertension and dyslipidemia (Barlow, 2007).

    In contrast with waist circumference, BMI and its associated disease and mortality risk appear
    to vary among ethnic subgroups. Female African American populations appear to have the lowest
    mortality risk at a BMI of 26.2-28.5 kg/m2 and 27.1-30.2 kg/m2 for women and men, respectively.
    In contrast, Asian populations may experience lowest mortality rates starting at a BMI of 23 to
    24 kg/m2. The correlation between BMI and diabetes risk also varies by ethnicity (LeBlanc,
    2011. p.2-3)

    Screening for BMI and follow-up therefore is critical to closing this gap and contributes to
    quality goals of population health and cost reduction. However, due to concerns for other
    underlying conditions (such as bone health) or nutrition related deficiencies providers are
    cautioned to use clinical judgment and  take these into account when considering weight
    management programs for overweight patients, especially the elderly (NHLBI Obesity Education
    Initiative, 1998, p. 91).

    BMI below Normal Parameters

    On the other end of the body weight spectrum is underweight (BMI <18.5 kg/m2), which is equally
    detrimental to population health. When compared to normal weight individuals(BMI 18.5-25
    kg/m2), underweight individuals have significantly higher death rates with a Hazard Ratio of
    2.27 and  95% confidence intervals (CI) = 1.78, 2.90 (Borrell & Lalitha (2014).

    Poor nutrition or underlying health conditions can result in underweight (Fryer & Ogden, 2012).
    The National Health and Nutrition Examination Survey (NHANES) results from the 2007-2010
    indicate that women are more likely to be underweight than men (2012). Therefore patients
    should be equally screened for underweight and followed up with nutritional counselling to
    reduce mortality and morbidity associated with underweight.

    Guidance: *  There is no diagnosis associated with this measure.
    *  This measure is to be reported a minimum of once per reporting period for patients seen
    during the reporting period.
    *  This measure may be reported by eligible professionals who perform the quality actions
    described in the measure based on the services provided at the time of the qualifying visit and
    the measure-specific denominator coding.

    BMI Measurement Guidance:
    *  Height and Weight - An eligible professional or their staff is required to measure both
    height and weight. Both height and weight must be measured within twelve months of the current
    encounter and may be obtained from separate encounters.  Self-reported values cannot be used.
    *  The BMI may be documented in the medical record of the provider or in outside medical
    records obtained by the provider.
    *  If the most recent documented BMI is outside of normal parameters, then a follow-up plan is
    documented during the encounter or during the previous twelve months of the current encounter.
    * If more than one BMI is reported during the measurement period, the most recent BMI will be
    used to determine if the performance has been met.
    * Review the exclusions criteria to determine those patients that BMI measurement may not be
    appropriate or necessary.

    Follow-Up Plan Guidance:

         1.      *  The documented follow-up plan must be based on the most recent documented BMI,
    outside of normal parameters, example: "Patient referred to nutrition counseling for BMI above
    or below normal parameters."

    (See Definitions for examples of follow-up plan treatments).
    Variation has been noted in studies exploring optimal BMI ranges for the elderly (see Donini et
    al., (2012); Holme and Tonstad (2015); and Diehr et al. (2008). Notably however, all these
    studies have arrived at ranges that differ from the standard range for ages 18 and older, which
    is >=18.5 and < 25 kg/m2. For instance, both Donini et al. (2012) and Holme and Tonstad (2015)
    reported findings that suggest that higher BMI (higher than the upper end of 25kg/m2) in the
    elderly may be beneficial. Similarly, worse outcomes have been associated with being
    underweight (at a threshold higher than 18.5 kg/m2) at age 65 (Diehr et al. 2008). Because of
    optimal BMI range variation recommendations from these studies, no specific optimal BMI range
    for the elderly is used. However, It may be appropriate to exempt certain patients from a
    follow-up plan by applying the exception criteria. Review the following to apply the Medical
    Reason exception criteria:
    The Medical Reason exception could include, but is not limited to, the following patients as
    deemed appropriate by the health care provider:
         *     Elderly Patients (65 or older) for whom weight reduction/weight gain would
    complicate other underlying health conditions such as the following examples:
                 *Illness or physical disability
                 *Mental illness, dementia, confusion
    *Nutritional deficiency such as Vitamin/mineral deficiency*

         *     Patients in an urgent or emergent medical situation where time is of the essence and
    to delay treatment would jeopardize the patient's health status

    More information: https://ecqi.healthit.gov/ecqm/measures/cms69v6
    """

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'Quality Insights',
    ]

    references = [
        'Barlow SE, the Expert Committee. Expert committee recommendations regarding the prevention, assessment, and treatment of child and adolescent overweight and obesity: summary report. Pediatrics 2007;120:S164-92.',
        'Barnes PM, & Schoenborn CA (2012). Trends in adults receiving a recommendation for exercise or other physical activity from a physician or other health professional. Centers for Disease Control and Prevention (CDC), National Center for Health Statistics (NCHS) Data Brief, No. 86: Feb 2012.',
        'Borrell, L.N. & Samuel, L. (2014). Body mass index categories and mortality risk in US adults: The effect of overweight and obesity on advancing death. American Journal of Public Health, 104, 512-519.',
        'Cawley J and Meyerhoefer C. The medical care costs of obesity: an instrumental variables approach. Journal of Health Economics, 31(1): 219-230, 2012.',
        'Diehr P, O\'Meara ES, Fitzpatrick A, Newman AB, Kuller L, Burke G. (2008) Weight, mortality, years of healthy life, and active life expectancy in older adults. Journal of American Geriatrics Society, 56, 76-83.',
        'Donini, L. M., Savina, C., Gennaro, E., De Felice, M. R., Rosano, A., Pandolfo, M. M., ... Chumlea, W. C. (2012). A Systematic Review Of The Literature Concerning The Relationship Between Obesity And Mortality In The Elderly. The Journal of Nutrition, Health & Aging, 16(1), 89-98.',
        'Finkelstein, E.A., Trogdon, J.G., Cohen, J.W., & Dietz, W. (2009). Annual Medical Spending Attributable To Obesity: Payer-And Service-Specific Estimates. Health Affairs, 28(5), w822-w831. doi: 10.1377/hlthaff.28.5.w822',
        'Fitch A, Everling L, Fox C, Goldberg J, Heim C, Johnson K, Kaufman T, Kennedy E, Kestenbaun C, Lano M, Leslie D, Newell T, O\'Connor P, Slusarek B, Spaniol A, Stovitz S, Webb B. Institute for Clinical Systems Improvement. Prevention and Management of Obesity for Adults, 4-10. Updated May 2013.',
        'Fryar, C. D., & Ogden, C. L. (2012). Prevalence of underweight among adults aged 20 and over: United States, 1960-1962 through 2007-2010. National Center for Health Statistics, Division of Health and Nutrition Examination Surveys. Retrieved from http://www.cdc.gov/nchs/data/hestat/underweight_adult_07_10/underweight_adult_07_10.pdf',
        'Garvey, W. T., Mechanick, J. I., Brett, E. M., Garber, A. J., Hurley, D. L., Jastreboff, A. M., Nadolsky, K., ... Reviewers of the AACE/ACE Obesity Clinical Practice Guidelines. (January 01, 2016). AMERICAN ASSOCIATION OF CLINICAL ENDOCRINOLOGISTS AND AMERICAN COLLEGE OF ENDOCRINOLOGY COMPREHENSIVE CLINICAL PRACTICE GUIDELINES FOR MEDICAL CARE OF PATIENTS WITH OBESITY. Endocrine Practice: Official Journal of the American College of Endocrinology and the American Association of Clinical Endocrinologists, 22, 1-203.',
        'Holme, I., & Tonstad, S. (2015) Survival in elderly men in relation to midlife and current BMI. Age and Ageing, 44, 3, 434-9.',
        'Jensen, M.D., Ryan, D.H., Apovian, C.M., Ard, J.D., Comuzzie, A. G., Donato, K.A., ... Yanovski, S.Z. (2013). Practice guidelines and the obesity society report of the american college of cardiology/american heart association task force on 2013 AHA/ACC/TOS guideline for the management of overweight and obesity in adults: A report of the american college of cardiology/american heart association task force on practice guidelines and the obesity society. Circulation. doi: 10.1161/01.cir.0000437739.71477.',
        'LeBlanc E, O\'Connor E, Whitlock EP, et al. Screening for and management of obesity and overweight in adults. Evidence Report No. 89. AHRQ Publication No. 11-05159-EF-1. Rockville, MD: Agency for Healthcare Research and Quality. October 2011.',
        'NHLBI Obesity Education Initiative. (1998). Clinical guidelines on the identification, evaluation, and treatment of overweight and obesity in adults.',
        'Ogden, C. L., Carroll, M. D., Kit, B. K., & Flegal, K. M. (2013). Prevalence of obesity among adults: United States, 2011-2012, Centers for Disease Control and Prevention (CDC), National Center for Health Statistics (NCHS) Data Brief, No. 131: Oct 2013. Retrieved from http://www.cdc.gov/nchs/data/databriefs/db131.pdf',
        'The State of Obesity. (n.d). Obesity Rates and Trends Overview. Retrieved from http://stateofobesity.org/obesity-rates-trends-overview/ (Access October 2016)',
        'The State of Obesity. (n.d). The Healthcare Costs of Obesity. Retrieved from http://stateofobesity.org/healthcare-costs-obesity/(Accessed October 2016)',
        'Wilkinson, J., Bass, C., Diem, S., Gravley, A., Harvey, L. Hayes, R., Johnson, K., Maciosek, M., McKeon, K., Milteer, L., Morgan, J., Rothe, P., Snellman, L., Solberg, L., Storlie, C., & Vincent, P. (2013). Institute for Clinical Systems Improvement. Preventive Services for Adults. Retrieved from https://www.icsi.org/_asset/gtjr9h/PrevServAdults-Interactive0912.pdf.',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: All patients 18 and older on the date of the encounter with at least
        one eligible encounter during the measurement period
        """
        age = self.patient.age_at(self.timeframe.end)
        if age < 18:
            return False
        return True

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Patients who are pregnant
        Patients receiving palliative care
        Patients who refuse measurement of height and/or weight or refuse follow-up

        Exceptions: Patients with a documented Medical Reason:
        * Elderly Patients (65 or older) for whom weight reduction/weight gain would complicate
        other underlying health conditions such as the following examples:
             *Illness or physical disability
             *Mental illness, dementia, confusion
             *Nutritional deficiency, such as Vitamin/mineral deficiency

        * Patients in an urgent or emergent medical situation where time is of the essence and to
        delay treatment would jeopardize the patient's health status
        """
        if not self.in_initial_population():
            return False

        # FIX: this isn't quite right - we need a more accurate way to
        # assess whether the patient is presently pregnant here
        if self.patient.conditions.within_one_year(self.timeframe).find(PregnancyDx):
            return False

        # FIX: how do we check for a PalliativeCareEncounter?  We have
        # Encounter but it doesn't appear to have codes attached.

        # FIX: how do we check to see if the patient has refused a
        # height/weight measure or refused a follow-up plan?

        return True

    def in_numerator(self):
        """
        Numerator: Patients with a documented BMI during the encounter or during the previous
        twelve months, AND when the BMI is outside of normal parameters, a follow-up plan is
        documented during the encounter or during the previous twelve months of the current
        encounter

        Exclusions: Not Applicable
        """
        pass

    def compute_results(self):
        """
        Clinical recommendation: As cited in Fetch et al. (2013), The Institute for Clinical
        Systems Improvement (ICSI) Health Care Guideline,  Prevention and Management of Obesity for
        Adults provides the Strength of Recommendation as Strong for the following:
               -Record height, weight and calculate body mass index at least annually
        -Clinicians should consider waist circumference measurement to estimate disease risk for
        patients who have normal or overweight BMI scores. For adult patients with a BMI of 25 to
        34.9 kg/m2, sex-specific waist circumference cutoffs should be used in conjunction with BMI
        to identify increased disease risk.

        Individuals who are overweight (BMI 25<30), and who do not have indicators of increased CVD
        risk (e.g., diabetes, pre-diabetes, hypertension, dyslipidemia, elevated waist
        circumference) or other obesity-related comorbidities and
        individuals who have a history of overweight and are now normal weight with risk factors at
        acceptable levels:

        'Advise to frequently measure their own weight, and to avoid weight gain by adjusting their
        food intake if they start to gain more than a few pounds. Also, advice patients that
        engaging in regular physical activity will help them avoid weight gain.' (2013 AHA/AAC/TOS
        Obesity Guideline, p. 20)

        'Advise overweight and obese individuals who would benefit from weight loss to participate
        for >=6 months in a comprehensive lifestyle program that assists participants in adhering
        to a lower calorie diet and in increasing physical activity through the use of behavioral
        strategies... NHLBI Grade A (Strong)' (2013 AHA/AAC/TOS Obesity Guideline, p. 15)

        USPSTF Clinical Guideline (Grade B Recommendation)
        Individuals with a body mass index (BMI) of 30 kg/m2 or higher should be offered or
        referred to intensive, multicomponent behavioral interventions that include the following
        components:
        -       Behavioral management activities, such as setting weight-loss goals
        -       Improving diet or nutrition and increasing physical activity
        -       Addressing barriers to change
        -       Self-monitoring
        -       Strategizing how to maintain lifestyle changes


        Nutritional safety for the elderly should be considered when recommending weight reduction.
        'A clinical decision to forego obesity treatment in older adults should be guided by an
        evaluation of the potential benefits of weight reduction for day-to-day functioning and
        reduction of the risk of future cardiovascular events, as well as the patient's motivation
        for weight reduction. Care must be taken to ensure that any weight reduction program
        minimizes the likelihood of adverse effects on bone health or other aspects of nutritional
        status' Evidence Category D. (NHLBI Obesity Education Initiative, 1998, p. 91). In
        addition, weight reduction prescriptions in older persons should be accompanied by proper
        nutritional counseling and regular body weight monitoring. (NHLBI Obesity Education
        Initiative, 1998, p. 91).

        The possibility that a standard approach to weight loss will work differently in diverse
        patient populations must be considered when setting expectations about treatment outcomes.
        Evidence Category B. (NHLBI Obesity Education Initiative, 1998).
        """
        recommendations: List[Recommendation] = []
        status = 'not_applicable'
        narrative = ''

        patient = self.patient
        bmi = self.compute_bmi()

        # FIX: the text discuses recommending a waist measurement for
        # certain BMI ranges, but doesn't say what to do with that
        # information and it's not in the numerator.  Include it?

        year_range = Timeframe(start=self.timeframe.end.shift(years=-1), end=self.timeframe.end)
        followups = patient.instructions.within(year_range).find(AboveNormalFollowUp |
                                                                 BelowNormalFollowUp)

        if not bmi:
            narrative = f"{patient.first_name} should have their height and weight measured to assess their BMI."
            status = 'due'
            recommendations.append(VitalSignRecommendation(patient=self.patient))

        elif (bmi >= 18.5 and bmi < 25.0):
            narrative = f"{patient.first_name}'s BMI {bmi:.1f} is within normal range. " \
                        f'This protocol is satisfied.'
            status = 'current'
        elif bmi < 18.5:
            if followups:
                status = 'current'
                narrative = f"{patient.first_name}'s BMI {bmi:.1f} is below the normal range (18.5-25). " \
                            f"A followup instruction given with the last year satisfies this protocol."
            else:
                recommendations.append(
                    InstructionRecommendation(
                        patient=self.patient, instruction=AboveNormalFollowUp))
                status = 'due'
                narrative = f"{patient.first_name}'s BMI {bmi:.1f} is below the normal range (18.5-25). " \
                            f"You should discuss a follow-up plan with {patient.first_name}."
        else:
            if followups:
                status = 'current'
                narrative = f"{patient.first_name}'s BMI {bmi:.1f} is above the normal range (18.5-25). " \
                            f"A followup instruction given with the last year satisfies this protocol."
            else:
                recommendations.append(
                    InstructionRecommendation(
                        patient=self.patient, instruction=AboveNormalFollowUp))
                status = 'due'
                narrative = f"{patient.first_name}'s BMI {bmi:.1f} is above the normal range (18.5-25). " \
                            f"You should discuss a follow-up plan with {patient.first_name}."

        return {'status': status, 'narrative': narrative, 'recommendations': recommendations}

    def compute_bmi(self) -> Optional[float]:
        patient = self.patient
        t = self.timeframe
        bmi = None

        # look for a BMI or a height and weight within the period
        bmi_v = patient.vitalsigns.within_one_year(t).find(BmiPercentile | BmiLoincValue).last()

        if bmi_v:
            bmi = float(bmi_v['value'])
        else:
            height_v = patient.vitalsigns.within_one_year(t).find(Height).last()
            weight_v = patient.vitalsigns.within_one_year(t).find(Weight).last()

            if (height_v and weight_v):
                height: Optional[float]
                weight: Optional[float]

                # convert to metric if needed
                if height_v['units'] == 'in':
                    height = float(height_v['value']) * 0.0254
                elif height_v['units'] == 'm':
                    height = float(height_v['value'])
                elif height_v['units'] == 'cm':
                    height = float(height_v['value']) / 100.0
                else:
                    # unexepected unit, ignore - could throw an error here but who would catch it?
                    height = None

                # convert to metric if needed
                if weight_v['units'] == 'lbs':
                    weight = float(weight_v['value']) * 0.453592
                elif weight_v['units'] == 'kg':
                    weight = float(weight_v['value'])
                else:
                    # unexepected unit, ignore - could throw an error here but who would catch it?
                    weight = None

                if height and weight:
                    bmi = weight / (height * height)

        return bmi



import arrow

from canvas_workflow_kit import settings

from canvas_workflow_kit.builtin_cqms.cms123v6_diabetes_foot_exam import ClinicalQualityMeasure123v6
from canvas_workflow_kit.patient_recordset import (
    BillingLineItemRecordSet,
    ConditionRecordSet,
    ProtocolOverrideRecordSet
)
from canvas_workflow_kit.tests.base import SDKBaseTest
from canvas_workflow_kit.timeframe import Timeframe


class ClinicalQualityMeasure123v6Test(SDKBaseTest):

    def setUp(self):
        self.mocks_path = f'{settings.BASE_DIR}/builtin_cqms/tests/mock_data'

    def test_description(self):
        expected = ('Patients 18-75 years of age with diabetes who '
                    'have not received a foot exam in the last year.')
        self.assertEqual(expected, ClinicalQualityMeasure123v6._meta.description)

    def test_information(self):
        expected = 'https://ecqi.healthit.gov/sites/default/files/ecqm/measures/CMS123v6.html'
        self.assertEqual(expected, ClinicalQualityMeasure123v6._meta.information)

    def test_enabled(self):
        self.assertTrue(ClinicalQualityMeasure123v6.enabled())

    def test_version(self):
        self.assertTrue(hasattr(ClinicalQualityMeasure123v6._meta, 'version'))

    def test_change_types(self):
        result = ClinicalQualityMeasure123v6._meta.compute_on_change_types
        expected = [
            'protocol_override',
            'condition',
            'interview',
            'patient',
        ]
        self.assertEqual(len(expected), len(result))
        for change in expected:
            self.assertIn(change, result)

    def test_period(self):
        start = arrow.get('2017-01-23 13:24:56')
        end = arrow.get('2018-01-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        patient = self.load_patient('cms123v6_diabetes_no')
        tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe)

        # no dates --> N/A
        self.assertEqual('N/A', tested.period)

        # identical dates
        tested._on_dates = []
        tested._on_dates.append(arrow.get('2018-08-23'))
        tested._on_dates.append(arrow.get('2018-08-23'))
        tested._on_dates.append(arrow.get('2018-08-23'))
        self.assertRegex(tested.period, r'^[\w\s]+ on 8/23/18$')

        # different dates
        tested._on_dates.append(arrow.get('2018-08-21'))
        self.assertEqual('between 8/21/18 and 8/23/18', tested.period)
        tested._on_dates.append(arrow.get('2018-08-22'))
        self.assertEqual('between 8/21/18 and 8/23/18', tested.period)
        tested._on_dates.append(arrow.get('2018-08-24'))
        self.assertEqual('between 8/21/18 and 8/24/18', tested.period)

    def test_in_denominator(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient has no diabetes --> false
        patient = self.load_patient('cms123v6_diabetes_no')
        tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

        # patient has diabetes --> true
        patient = self.load_patient('cms123v6_diabetes_yes')
        tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

        # -- patient to young < 18 --> false
        patient = self.load_patient('cms123v6_diabetes_yes')
        patient.patient['birthDate'] = '2000-08-24'
        self.assertTrue(17 < patient.age_at(timeframe.end) < 18)
        tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())
        patient.patient['birthDate'] = '2000-08-23'
        self.assertTrue(patient.age_at(timeframe.end) == 18)
        tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

        # -- patient to old > 75 --> false
        patient = self.load_patient('cms123v6_diabetes_yes')
        patient.patient['birthDate'] = '1943-08-22'
        self.assertTrue(75 < patient.age_at(timeframe.end) < 76)
        tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())
        patient.patient['birthDate'] = '1943-08-23'
        self.assertTrue(patient.age_at(timeframe.end) == 75)
        tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())
        patient.patient['birthDate'] = '1943-08-24'
        self.assertTrue(patient.age_at(timeframe.end) < 75)
        tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

        # -- patient has an other code --> false
        patient = self.load_patient('cms123v6_diabetes_yes')
        for condition in patient.conditions:
            for code in condition['coding']:
                code['code'] = 'code'
        tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

        # -- patient had amputations --> false
        tests = [
            ['Q7223'],  # bilateral amputation
            ['Q7222', 'Q7220'],  # left amputation
            ['Q7221', 'Q7222'],  # right amputation
            ['Q7220', 'Q7221'],  # unspecified amputation
        ]
        for codes in tests:
            patient = self.load_patient('cms123v6_diabetes_yes')
            amputations = []
            for code in codes:
                amputations.append({
                    'clinicalStatus': 'active',
                    'coding': [{
                        'system': 'ICD-10',
                        'code': code
                    }],
                    'periods': [{
                        'from': '1998-08-22',
                        'to': None
                    }]
                })
            patient.conditions += ConditionRecordSet(amputations)
            tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe)
            self.assertFalse(tested.in_denominator(), 'for the code %s' % code)

        # patient has diabetes + already perform the exam --> true
        patient = self.load_patient('cms123v6_diabetes_yesdone')
        tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

    def test_in_numerator(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # not test when patient has no diabetes because it makes no sense
        # patient = self.load_patient('cms123v6_diabetes_no')

        # patient has diabetes + no exam performed --> false
        patient = self.load_patient('cms123v6_diabetes_yes')
        tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())

        # patient has diabetes + already perform the exam --> true
        patient = self.load_patient('cms123v6_diabetes_yesdone')
        tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())

        # -- all exams are required
        tests = ['134388005', '91161007', '401191002']
        for test in tests:
            patient = self.load_patient('cms123v6_diabetes_yesdone')
            for interview in patient.interviews:
                records = interview['questions']
                interview['questions'] = []
                for record in records:
                    if record['code'] != test:
                        interview['questions'].append(record)
            tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe)
            self.assertFalse(tested.in_numerator())

        # -- exams have to have been performed within the year
        patient = self.load_patient('cms123v6_diabetes_yesdone')
        for interview in patient.interviews:
            interview['noteTimestamp'] = '2017-08-23 13:24:55'
        tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())
        for interview in patient.interviews:
            interview['noteTimestamp'] = '2017-08-23 13:24:57'
        tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())

        # -- exams have to be committed
        patient = self.load_patient('cms123v6_diabetes_yesdone')
        for interview in patient.interviews:
            interview['committer'] = None  # id of the person
        tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())

        # -- exams have to be not entered in error
        patient = self.load_patient('cms123v6_diabetes_yesdone')
        for interview in patient.interviews:
            interview['entered_in_error'] = 3  # id of the person
        tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())

        # -- exams have to be not deleted
        patient = self.load_patient('cms123v6_diabetes_yesdone')
        for interview in patient.interviews:
            interview['deleted'] = True
        tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())

    def test_in_numerator_with_override(self):
        patient = self.load_patient('cms123v6_diabetes_yesdone')
        patient.protocol_overrides = ProtocolOverrideRecordSet([{
            'protocolKey': 'ClinicalQualityMeasure123v6',
            'adjustment': {
                'reference': '2018-02-01T00:00:00Z',
                'cycleDays': 180
            },
            'snooze': None,
            'modified': '2018-03-15T12:32:54Z',
        }])  # yapf: disable
        patient.interviews[0]['noteTimestamp'] = '2018-08-05'

        tested = ClinicalQualityMeasure123v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=0))
        self.assertTrue(tested.in_numerator())

        tested = ClinicalQualityMeasure123v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=80))
        self.assertTrue(tested.in_numerator())

        tested = ClinicalQualityMeasure123v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=180))
        self.assertTrue(tested.in_numerator())

        tested = ClinicalQualityMeasure123v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=181))
        self.assertFalse(tested.in_numerator())

    def test_compute_results(self):
        start = arrow.get('2017-10-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        now = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient has no diabetes --> not in denominator and numerator
        patient = self.load_patient('cms123v6_diabetes_no')
        tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertIsNotApplicable(result)

        # patient has diabetes --> in denominator, not numerator
        patient = self.load_patient('cms123v6_diabetes_yes')
        tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        text = 'Dohav has diabetes and is due for foot exam.'
        self.assertEqual(text, result.narrative)
        expected = [
            {
                'title': (
                    'Conduct comprehensive foot examination '
                    'including assessment of protective sensation, pulses and visual inspection.'),
                'narrative': None,
                'command': {
                    'type': 'interview',
                    'filter': {
                        'coding': [{
                            'system': 'snomedct',
                            'code': ['401191002', '91161007', '134388005']
                        }]
                    }
                },
                'key': 'CMS123v6_RECOMMEND_FOOT_EXAM',
                'rank': 1,
                'button': 'Plan',
            },
        ]
        for exp, recommendation in zip(expected, result.recommendations):
            self.assertEqual(exp['key'], recommendation.key)
            self.assertEqual(exp['rank'], recommendation.rank)
            self.assertEqual(exp['button'], recommendation.button)
            self.assertEqual(exp['title'], recommendation.title)
            self.assertEqual(exp['narrative'], recommendation.narrative)
            self.assertEqual(exp['command']['type'], recommendation.command['type'])
            exp_item = exp['command']['filter']['coding']
            rec_item = recommendation.command['filter']['coding']
            self.assertEqual(len(exp_item), len(rec_item))
            for exp_coding, rec_coding in zip(exp_item, rec_item):
                self.assertEqual(exp_coding['system'], rec_coding['system'])
                self.assertEqual(exp_coding['code'], rec_coding['code'])
        self.assertEqual(-1, result.due_in)

        # patient has diabetes + already perform the exam --> in denominator and numerator
        patient = self.load_patient('cms123v6_diabetes_yesdone')
        tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('satisfied', result.status)
        text = r'^Dohav has diabetes and his comprehensive foot exam was done [\w\s]+ on 8/22/18.$'
        self.assertRegex(result.narrative, text)
        self.assertEqual([], result.recommendations)
        self.assertEqual(303, result.due_in)  # not one year since the time frame define the period

        # patient has diabetes + younger than 18 --> n/a + due in
        patient = self.load_patient('cms123v6_diabetes_no')
        patient.patient['birthDate'] = '2000-08-30'
        tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('not_applicable', result.status)
        self.assertIsNone(result.due_in)
        patient = self.load_patient('cms123v6_diabetes_yes')
        patient.patient['birthDate'] = '2000-08-30'
        tested = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe, now=now)
        result = tested.compute_results()
        self.assertEqual('not_applicable', result.status)
        self.assertEqual(6, result.due_in)

    def test_compute_results_with_override(self):
        patient = self.load_patient('cms123v6_diabetes_yesdone')
        patient.protocol_overrides = ProtocolOverrideRecordSet([{
            'protocolKey': 'ClinicalQualityMeasure123v6',
            'adjustment': {
                'reference': '2018-02-01T00:00:00Z',
                'cycleDays': 180
            },
            'snooze': None,
            'modified': '2018-03-15T12:32:54Z',
        }])  # yapf: disable
        patient.interviews[0]['noteTimestamp'] = '2018-08-05'

        # exam done within the last 180 days
        tested = ClinicalQualityMeasure123v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=80))
        result = tested.compute_results()
        self.assertEqual('satisfied', result.status)
        text = ('Dohav has diabetes and his comprehensive foot '
                'exam was done 2 months ago on 8/5/18.')
        self.assertEqual(text, result.narrative)
        self.assertEqual(0, len(result.recommendations))
        self.assertEqual(100, result.due_in)

        # exam done 181 days
        patient.billing_line_items = BillingLineItemRecordSet([{
            'id': 92,
            'created': arrow.get('2018-08-01').shift(days=181),
            'datetimeOfService': arrow.get('2018-08-01').shift(days=181),
            'cpt': '99202',
            'units': 1
        }])
        tested = ClinicalQualityMeasure123v6(
            patient=patient, now=arrow.get('2018-08-05').shift(days=181))
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        text = 'Dohav has diabetes and is due for foot exam.'
        self.assertEqual(text, result.narrative)
        self.assertEqual(1, len(result.recommendations))
        self.assertEqual(-1, result.due_in)
